<?php

namespace Drupal\sirv\Plugin\Field\FieldFormatter;

use Drupal\sirv\SirvUrlInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Url;
use Drupal\image\Plugin\Field\FieldFormatter\ImageFormatterBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'sirv_image' formatter.
 *
 * @FieldFormatter(
 *   id = "sirv_image",
 *   label = @Translation("Sirv image"),
 *   field_types = {
 *     "image"
 *   },
 *   weight = 10
 * )
 */
class SirvImageFormatter extends ImageFormatterBase implements ContainerFactoryPluginInterface {

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The Sirv URL sirvice.
   *
   * @var \Drupal\sirv\SirvUrlInterface
   */
  protected $sirvUrlService;

  /**
   * Constructs a SirvImageFormatter object.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\sirv\SirvUrlInterface $sirv_url_service
   *   The Sirv URL service.
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, ConfigFactoryInterface $config_factory, SirvUrlInterface $sirv_url_service) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);

    $this->configFactory = $config_factory;
    $this->sirvUrlService = $sirv_url_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('config.factory'),
      $container->get('sirv.url')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'image_style' => '',
      'sirv_profile' => '',
      'extra_sirv_options' => '',
      'image_link' => '',
      'responsive' => TRUE,
      'responsive_scale_crop' => '',
      'responsive_resize_step' => '',
      'lazy_load' => TRUE,
      'lazy_load_threshold' => '',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = [];

    $image_styles = image_style_options(FALSE);
    $element['image_style'] = [
      '#title' => $this->t('Image style'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('image_style'),
      '#empty_option' => $this->t('None (original image)'),
      '#options' => $image_styles,
      '#description' => $this->t('The original image is usually desired, though an image style may be selected if it provides an effect that is not available in Sirv, such as custom cropping.'),
    ];

    $element['sirv_profile'] = [
      '#title' => $this->t('Sirv profile'),
      '#type' => 'textfield',
      '#description' => $this->t('Enter a Sirv profile to use for processing images in this field.'),
      '#default_value' => $this->getSetting('sirv_profile'),
    ];
    $element['extra_sirv_options'] = [
      '#title' => $this->t('Extra Sirv options'),
      '#type' => 'textfield',
      '#description' => $this->t('These options will override options from the profile. Example of format: %example', ['%example' => 'scale.width=250&grayscale=true']),
      '#default_value' => $this->getSetting('extra_sirv_options'),
    ];

    $element['responsive'] = [
      '#title' => $this->t('Make responsive'),
      '#type' => 'checkbox',
      '#description' => $this->t('Check to enable responsive behavior for this field.'),
      '#default_value' => $this->getSetting('responsive'),
    ];
    $element['responsive_scale_crop'] = [
      '#title' => $this->t('Automatic scale/crop'),
      '#type' => 'select',
      '#description' => $this->t('Select an option for scaling and cropping an image to fit a container.'),
      '#default_value' => $this->getSetting('responsive_scale_crop'),
      '#options' => [
        'contain' => $this->t('Contain (default)'),
        'cover' => $this->t('Cover'),
        'crop' => $this->t('Crop'),
      ],
    ];
    $element['responsive_resize_step'] = [
      '#title' => $this->t('Resize step'),
      '#type' => 'number',
      '#description' => $this->t('Enter the change in browser width required for a new responsive image to be generated.'),
      '#default_value' => $this->getSetting('responsive_resize_step'),
      '#size' => 10,
      '#field_suffix' => $this->t('px'),
    ];

    $element['lazy_load'] = [
      '#title' => $this->t('Use lazy loading'),
      '#type' => 'checkbox',
      '#description' => $this->t('Check to enable lazy loading for this field.'),
      '#default_value' => $this->getSetting('lazy_load'),
    ];
    $element['lazy_load_threshold'] = [
      '#title' => $this->t('Threshold'),
      '#type' => 'textfield',
      '#description' => $this->t('Enter a threshold value to indicate when to lazy load the image before it comes into view.'),
      '#default_value' => $this->getSetting('lazy_load_threshold'),
      '#size' => 10,
    ];

    $link_types = [
      'content' => $this->t('Content'),
      'file' => $this->t('File'),
    ];
    $element['image_link'] = [
      '#title' => $this->t('Link image to'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('image_link'),
      '#empty_option' => $this->t('Nothing'),
      '#options' => $link_types,
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];

    $image_styles = image_style_options(FALSE);
    // Unset possible 'No defined styles' option.
    unset($image_styles['']);
    // Styles could be lost because of enabled/disabled modules that defines
    // their styles in code.
    $image_style_setting = $this->getSetting('image_style');

    if (isset($image_styles[$image_style_setting])) {
      $summary[] = t('Image style: @style', ['@style' => $image_styles[$image_style_setting]]);
    }
    else {
      $summary[] = $this->t('Original image');
    }

    if (!empty($this->getSetting('sirv_profile'))) {
      $summary[] = $this->t('Sirv profile: @profile', ['@profile' => $this->getSetting('sirv_profile')]);
    }

    if (!empty($this->getSetting('extra_sirv_options'))) {
      $summary[] = $this->t('Extra Sirv options');
    }

    if (!empty($this->getSetting('responsive'))) {
      $summary[] = $this->t('Responsive');
    }

    if (!empty($this->getSetting('lazy_load'))) {
      $summary[] = $this->t('Lazy load');
    }

    $link_types = [
      'content' => $this->t('Linked to content'),
      'file' => $this->t('Linked to file'),
    ];
    // Display this setting only if image is linked.
    if (isset($link_types[$this->getSetting('image_link')])) {
      $summary[] = $link_types[$this->getSetting('image_link')];
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = [];
    $files = $this->getEntitiesToView($items, $langcode);

    // Early opt-out if the field is empty.
    if (empty($files)) {
      return $elements;
    }

    $field_settings = $this->getSettings();

    // Check if the formatter involves a link.
    $url = NULL;
    if ($field_settings['image_link'] == 'content') {
      $entity = $items->getEntity();
      if (!$entity->isNew()) {
        $url = $entity->toUrl();
      }
    }
    elseif ($field_settings['image_link'] == 'file') {
      $link_file = TRUE;
    }

    $responsive = $this->getSetting('responsive');
    $lazy_load = $this->getSetting('lazy_load');

    foreach ($files as $delta => $file) {
      // Get the Sirv URL.
      $sirv_url = $this->sirvUrlService->generateFromFile($file, $field_settings);

      if (isset($link_file)) {
        $url = Url::fromUri(file_create_url($sirv_url));
      }
      $cache_tags = $file->getCacheTags();

      // Extract field item attributes for the theme function, and unset them
      // from the $item so that the field template does not re-render them.
      $item = $file->_referringItem;
      $item_attributes = $item->_attributes;
      unset($item->_attributes);

      // Make sure the "class" attribute exists.
      if (!isset($item_attributes['class'])) {
        $item_attributes['class'] = [];
      }

      $data_options = [];

      if ($responsive || $lazy_load) {
        $item_attributes['data-src'] = $sirv_url;
        $item_attributes['class'][] = 'Sirv';
      }

      if ($responsive) {
        if (!$lazy_load) {
          $data_options['lazy'] = 'false';
        }

        // Handle the Responsive Scale/Crop setting.
        switch ($field_settings['responsive_scale_crop']) {
          case 'contain':
            $item_attributes['class'][] = 'image-fit';
            break;

          case 'cover':
            $item_attributes['class'][] = 'image-fill';
            $data_options['fit'] = 'cover';
            break;

          case 'crop':
            $item_attributes['class'][] = 'image-fill';
            $data_options['fit'] = 'crop';
            break;
        }

        // Handle the Responsive Resize Step setting.
        if ($field_settings['responsive_resize_step']) {
          $data_options['resizeStep'] = $field_settings['responsive_resize_step'];
        }
      }

      if ($lazy_load) {
        if (!$responsive) {
          $data_options['resize'] = 'false';
        }

        // Handle the Lazy Load Threshold setting.
        if ($field_settings['lazy_load_threshold']) {
          $data_options['threshold'] = $field_settings['lazy_load_threshold'];
        }
      }

      $data_options_attribute = NULL;
      if (!empty($data_options)) {
        $data_options_items = [];
        foreach ($data_options as $key => $value) {
          $data_options_items[] = "$key:$value";
        }
        $data_options_attribute = implode(';', $data_options_items);
      }

      if ($data_options_attribute) {
        $item_attributes['data-options'] = $data_options_attribute;
      }

      $image = [
        '#theme' => 'image',
        '#attributes' => $item_attributes,
        '#alt' => $item->alt,
      ];

      // Do not output an empty 'title' attribute.
      if (mb_strlen($item->title) != 0) {
        $image['#title'] = $item->title;
      }

      if (isset($sirv_url) && !$responsive && !$lazy_load) {
        $image['#uri'] = $sirv_url;
      }

      $elements[$delta] = [
        '#theme' => 'sirv_image',
        '#item' => $item,
        '#image' => $image,
        '#image_style' => $field_settings['image_style'],
        '#url' => $url,
        '#sirv' => [
          'url' => $sirv_url,
          'profile' => $field_settings['sirv_profile'],
          'responsive' => $responsive,
          'lazy_load' => $lazy_load,
        ],
        '#cache' => [
          'tags' => $cache_tags,
        ],
      ];

      // Add the Sirv library if required.
      if ($responsive || $lazy_load) {
        $elements[$delta]['#attached']['library'] = 'sirv/sirv-nospin';
      }
    }

    return $elements;
  }

}
